@@ -10,13 +10,13 @@ import (
 	"time"
 
 	"github.com/containers/common/pkg/completion"
-	"github.com/containers/podman/v4/cmd/podman/common"
-	"github.com/containers/podman/v4/cmd/podman/registry"
-	"github.com/containers/podman/v4/cmd/podman/validate"
-	"github.com/containers/podman/v4/libpod/define"
-	"github.com/containers/podman/v4/pkg/domain/entities"
-	envLib "github.com/containers/podman/v4/pkg/env"
-	"github.com/containers/podman/v4/pkg/rootless"
+	"github.com/containers/podman/v5/cmd/podman/common"
+	"github.com/containers/podman/v5/cmd/podman/registry"
+	"github.com/containers/podman/v5/cmd/podman/validate"
+	"github.com/containers/podman/v5/libpod/define"
+	"github.com/containers/podman/v5/pkg/domain/entities"
+	envLib "github.com/containers/podman/v5/pkg/env"
+	"github.com/containers/podman/v5/pkg/rootless"
 	"github.com/spf13/cobra"
 )
 
@@ -24,7 +24,7 @@ var (
 	execDescription = `Execute the specified command inside a running container.
 `
 	execCommand = &cobra.Command{
-		Use:               "exec [options] CONTAINER [COMMAND [ARG...]]",
+		Use:               "exec [options] CONTAINER COMMAND [ARG...]",
 		Short:             "Run a process in a running container",
 		Long:              execDescription,
 		RunE:              exec,
@@ -50,9 +50,11 @@ var (
 	envInput, envFile []string
 	execOpts          entities.ExecOptions
 	execDetach        bool
+	execCidFile       string
 )
 
 func execFlags(cmd *cobra.Command) {
+	podmanConfig := registry.PodmanConfig()
 	flags := cmd.Flags()
 
 	flags.SetInterspersed(false)
@@ -62,16 +64,20 @@ func execFlags(cmd *cobra.Command) {
 	flags.StringVar(&execOpts.DetachKeys, detachKeysFlagName, containerConfig.DetachKeys(), "Select the key sequence for detaching a container. Format is a single character [a-Z] or ctrl-<value> where <value> is one of: a-z, @, ^, [, , or _")
 	_ = cmd.RegisterFlagCompletionFunc(detachKeysFlagName, common.AutocompleteDetachKeys)
 
+	cidfileFlagName := "cidfile"
+	flags.StringVar(&execCidFile, cidfileFlagName, "", "File to read the container ID from")
+	_ = cmd.RegisterFlagCompletionFunc(cidfileFlagName, completion.AutocompleteDefault)
+
 	envFlagName := "env"
 	flags.StringArrayVarP(&envInput, envFlagName, "e", []string{}, "Set environment variables")
 	_ = cmd.RegisterFlagCompletionFunc(envFlagName, completion.AutocompleteNone)
 
 	envFileFlagName := "env-file"
-	flags.StringSliceVar(&envFile, envFileFlagName, []string{}, "Read in a file of environment variables")
+	flags.StringArrayVar(&envFile, envFileFlagName, []string{}, "Read in a file of environment variables")
 	_ = cmd.RegisterFlagCompletionFunc(envFileFlagName, completion.AutocompleteDefault)
 
-	flags.BoolVarP(&execOpts.Interactive, "interactive", "i", false, "Keep STDIN open even if not attached")
-	flags.BoolVar(&execOpts.Privileged, "privileged", false, "Give the process extended Linux capabilities inside the container.  The default is false")
+	flags.BoolVarP(&execOpts.Interactive, "interactive", "i", false, "Make STDIN available to the contained process")
+	flags.BoolVar(&execOpts.Privileged, "privileged", podmanConfig.ContainersConfDefaultsRO.Containers.Privileged, "Give the process extended Linux capabilities inside the container.  The default is false")
 	flags.BoolVarP(&execOpts.Tty, "tty", "t", false, "Allocate a pseudo-TTY. The default is false")
 
 	userFlagName := "user"
@@ -82,6 +88,10 @@ func execFlags(cmd *cobra.Command) {
 	flags.UintVar(&execOpts.PreserveFDs, preserveFdsFlagName, 0, "Pass N additional file descriptors to the container")
 	_ = cmd.RegisterFlagCompletionFunc(preserveFdsFlagName, completion.AutocompleteNone)
 
+	preserveFdFlagName := "preserve-fd"
+	flags.UintSliceVar(&execOpts.PreserveFD, preserveFdFlagName, nil, "Pass a list of additional file descriptors to the container")
+	_ = cmd.RegisterFlagCompletionFunc(preserveFdFlagName, completion.AutocompleteNone)
+
 	workdirFlagName := "workdir"
 	flags.StringVarP(&execOpts.WorkDir, workdirFlagName, "w", "", "Working directory inside the container")
 	_ = cmd.RegisterFlagCompletionFunc(workdirFlagName, completion.AutocompleteDefault)
@@ -111,16 +121,12 @@ func init() {
 }
 
 func exec(cmd *cobra.Command, args []string) error {
-	var nameOrID string
+	nameOrID, command, err := determineTargetCtrAndCmd(args, execOpts.Latest, execCidFile != "")
+	if err != nil {
+		return err
+	}
+	execOpts.Cmd = command
 
-	if len(args) == 0 && !execOpts.Latest {
-		return errors.New("exec requires the name or ID of a container or the --latest flag")
-	}
-	execOpts.Cmd = args
-	if !execOpts.Latest {
-		execOpts.Cmd = args[1:]
-		nameOrID = strings.TrimPrefix(args[0], "/")
-	}
 	// Validate given environment variables
 	execOpts.Envs = make(map[string]string)
 	for _, f := range envFile {
@@ -138,6 +144,12 @@ func exec(cmd *cobra.Command, args []string) error {
 
 	execOpts.Envs = envLib.Join(execOpts.Envs, cliEnv)
 
+	for _, fd := range execOpts.PreserveFD {
+		if !rootless.IsFdInherited(int(fd)) {
+			return fmt.Errorf("file descriptor %d is not available - the preserve-fd option requires that file descriptors must be passed", fd)
+		}
+	}
+
 	for fd := 3; fd < int(3+execOpts.PreserveFDs); fd++ {
 		if !rootless.IsFdInherited(fd) {
 			return fmt.Errorf("file descriptor %d is not available - the preserve-fds option requires that file descriptors must be passed", fd)
@@ -167,12 +179,12 @@ func exec(cmd *cobra.Command, args []string) error {
 		streams.AttachOutput = true
 		streams.AttachError = true
 
-		exitCode, err := registry.ContainerEngine().ContainerExec(registry.GetContext(), nameOrID, execOpts, streams)
+		exitCode, err := registry.ContainerEngine().ContainerExec(registry.Context(), nameOrID, execOpts, streams)
 		registry.SetExitCode(exitCode)
 		return err
 	}
 
-	id, err := registry.ContainerEngine().ContainerExecDetached(registry.GetContext(), nameOrID, execOpts)
+	id, err := registry.ContainerEngine().ContainerExecDetached(registry.Context(), nameOrID, execOpts)
 	if err != nil {
 		return err
 	}
@@ -180,6 +192,33 @@ func exec(cmd *cobra.Command, args []string) error {
 	return nil
 }
 
+// determineTargetCtrAndCmd determines which command exec should run in which container
+func determineTargetCtrAndCmd(args []string, latestSpecified bool, execCidFileProvided bool) (string, []string, error) {
+	var nameOrID string
+	var command []string
+
+	if len(args) == 0 && !latestSpecified && !execCidFileProvided {
+		return "", nil, errors.New("exec requires the name or ID of a container or the --latest or --cidfile flag")
+	} else if latestSpecified && execCidFileProvided {
+		return "", nil, errors.New("--latest and --cidfile can not be used together")
+	}
+	command = args
+	if !latestSpecified {
+		if !execCidFileProvided {
+			// assume first arg to be name or ID
+			command = args[1:]
+			nameOrID = strings.TrimPrefix(args[0], "/")
+		} else {
+			content, err := os.ReadFile(execCidFile)
+			if err != nil {
+				return "", nil, fmt.Errorf("reading CIDFile: %w", err)
+			}
+			nameOrID = strings.Split(string(content), "\n")[0]
+		}
+	}
+	return nameOrID, command, nil
+}
+
 func execWait(ctr string, seconds int32) error {
 	maxDuration := time.Duration(seconds) * time.Second
 	interval := 100 * time.Millisecond
